/***************************************************************************
 * Copyright 2013 Kieker Project (http://kieker-monitoring.net)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ***************************************************************************/
package kieker.webgui.common;

import java.lang.annotation.Annotation;

import kieker.analysis.AnalysisController;
import kieker.analysis.AnalysisController.AnalysisControllerWithMapping;
import kieker.analysis.AnalysisControllerThread;
import kieker.analysis.display.HtmlText;
import kieker.analysis.display.Image;
import kieker.analysis.display.PlainText;
import kieker.analysis.display.annotation.Display;
import kieker.analysis.plugin.AbstractPlugin;
import kieker.analysis.plugin.annotation.InputPort;
import kieker.analysis.plugin.annotation.OutputPort;
import kieker.analysis.plugin.annotation.Plugin;
import kieker.analysis.plugin.annotation.Property;
import kieker.analysis.plugin.annotation.RepositoryPort;
import kieker.analysis.plugin.filter.AbstractFilterPlugin;
import kieker.analysis.plugin.reader.AbstractReaderPlugin;
import kieker.analysis.repository.AbstractRepository;
import kieker.analysis.repository.annotation.Repository;
import kieker.common.logging.LogImplWebguiLogging;
import kieker.webgui.common.exception.ReflectionException;

/**
 * The {@link ClassAndMethodContainer} is a container which contains - as the name already tells - various classes and methods. To be more precisely, it uses a
 * given class loader to load the equivalence of specific classes via reflection within this application to ensure that comparisons, assignments and the use of
 * specific methods will be done correctly. This is necessary as for every project within this application there will be a number of libraries which will be combined
 * in one class loader. This will result in multiple version of one and the same class and therefore in problems, if one doesn't use the correct class version.<br>
 * </br>
 * 
 * As we use the Mirror framework in this project (which simplifies some accesses to the Java reflection API) not all methods and classes have to be loaded in this
 * class.
 * 
 * @author Nils Christian Ehmke
 */
public class ClassAndMethodContainer {

	private Class<?> logImplWebguiLoggingClass;
	private Class<?> analysisControllerWithMappingClass;
	private Class<?> analysisControllerClass;
	private Class<?> analysisControllerThreadClass;
	private Class<?> abstractRepositoryClass;
	private Class<?> abstractPluginClass;
	private Class<?> abstractFilterPluginClass;
	private Class<?> abstractReaderPluginClass;
	private Class<?> imageClass;
	private Class<?> plainTextClass;
	private Class<?> htmlTextClass;

	private Class<? extends Annotation> pluginAnnotationClass;
	private Class<? extends Annotation> repositoryAnnotationClass;
	private Class<? extends Annotation> propertyAnnotationClass;
	private Class<? extends Annotation> outputPortAnnotationClass;
	private Class<? extends Annotation> inputPortAnnotationClass;
	private Class<? extends Annotation> repositoryPortAnnotationClass;
	private Class<? extends Annotation> displayAnnotationClass;

	/**
	 * Creates a new instance of this class, using the given class loader.
	 * 
	 * @param classLoader
	 *            The class loader which will be used to load all classes and determine their methods.
	 * 
	 * @throws ReflectionException
	 *             If one or more of the classes or methods for this container could not be found.
	 */
	@SuppressWarnings("unchecked")
	public ClassAndMethodContainer(final ClassLoader classLoader) throws ReflectionException {
		try {

			// For the first: Use the classloader the load all classes we will need.
			this.logImplWebguiLoggingClass = classLoader.loadClass(LogImplWebguiLogging.class.getName());

			this.analysisControllerWithMappingClass = classLoader.loadClass(AnalysisControllerWithMapping.class.getName());

			this.analysisControllerClass = classLoader.loadClass(AnalysisController.class.getName());
			this.analysisControllerThreadClass = classLoader.loadClass(AnalysisControllerThread.class.getName());
			this.abstractFilterPluginClass = classLoader.loadClass(AbstractFilterPlugin.class.getName());
			this.abstractReaderPluginClass = classLoader.loadClass(AbstractReaderPlugin.class.getName());
			this.abstractRepositoryClass = classLoader.loadClass(AbstractRepository.class.getName());
			this.abstractPluginClass = classLoader.loadClass(AbstractPlugin.class.getName());
			this.htmlTextClass = classLoader.loadClass(HtmlText.class.getName());
			this.plainTextClass = classLoader.loadClass(PlainText.class.getName());
			this.imageClass = classLoader.loadClass(Image.class.getName());

			// Now we load the more specific annotation classes
			this.pluginAnnotationClass = (Class<? extends Annotation>) classLoader.loadClass(Plugin.class.getName());
			this.repositoryAnnotationClass = (Class<? extends Annotation>) classLoader.loadClass(Repository.class.getName());
			this.propertyAnnotationClass = (Class<? extends Annotation>) classLoader.loadClass(Property.class.getName());
			this.outputPortAnnotationClass = (Class<? extends Annotation>) classLoader.loadClass(OutputPort.class.getName());
			this.inputPortAnnotationClass = (Class<? extends Annotation>) classLoader.loadClass(InputPort.class.getName());
			this.repositoryPortAnnotationClass = (Class<? extends Annotation>) classLoader.loadClass(RepositoryPort.class.getName());
			this.displayAnnotationClass = (Class<? extends Annotation>) classLoader.loadClass(Display.class.getName());

		} catch (final ClassNotFoundException ex) {
			// Something went wrong. We can do nothing except throwing an exception
			throw new ReflectionException("An error occured while loading the classes and methods.", ex);
		} catch (final SecurityException ex) {
			// Something went wrong. We can do nothing except throwing an exception
			throw new ReflectionException("An error occured while loading the classes and methods.", ex);
		} catch (final ClassCastException ex) {
			// Something went wrong. We can do nothing except throwing an exception
			throw new ReflectionException("An error occured while loading the classes and methods.", ex);
		}
	}

	public Class<?> getAbstractFilterPluginClass() {
		return this.abstractFilterPluginClass;
	}

	public Class<?> getAbstractReaderPluginClass() {
		return this.abstractReaderPluginClass;
	}

	public Class<? extends Annotation> getPluginAnnotationClass() {
		return this.pluginAnnotationClass;
	}

	public Class<? extends Annotation> getRepositoryAnnotationClass() {
		return this.repositoryAnnotationClass;
	}

	public Class<? extends Annotation> getPropertyAnnotationClass() {
		return this.propertyAnnotationClass;
	}

	public Class<? extends Annotation> getOutputPortAnnotationClass() {
		return this.outputPortAnnotationClass;
	}

	public Class<? extends Annotation> getInputPortAnnotationClass() {
		return this.inputPortAnnotationClass;
	}

	public Class<? extends Annotation> getRepositoryPortAnnotationClass() {
		return this.repositoryPortAnnotationClass;
	}

	public Class<? extends Annotation> getDisplayAnnotationClass() {
		return this.displayAnnotationClass;
	}

	public Class<?> getAbstractRepositoryClass() {
		return this.abstractRepositoryClass;
	}

	public Class<?> getAbstractPluginClass() {
		return this.abstractPluginClass;
	}

	public Class<?> getImageClass() {
		return this.imageClass;
	}

	public Class<?> getPlainTextClass() {
		return this.plainTextClass;
	}

	public Class<?> getHtmlTextClass() {
		return this.htmlTextClass;
	}

	public Class<?> getAnalysisControllerClass() {
		return this.analysisControllerClass;
	}

	public Class<?> getAnalysisControllerWithMappingClass() {
		return this.analysisControllerWithMappingClass;
	}

	public Class<?> getAnalysisControllerThreadClass() {
		return this.analysisControllerThreadClass;
	}

	public Class<?> getLogImplWebguiLoggingClass() {
		return this.logImplWebguiLoggingClass;
	}

}
