/***************************************************************************
 * Copyright 2013 Kieker Project (http://kieker-monitoring.net)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ***************************************************************************/

package kieker.webgui.service.impl.utility;

import java.io.File;
import java.lang.annotation.Annotation;
import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import kieker.common.logging.Log;
import kieker.common.logging.LogFactory;
import kieker.webgui.common.ClassAndMethodContainer;
import kieker.webgui.common.exception.AnalysisInitializationException;
import kieker.webgui.common.exception.InvalidAnalysisStateException;
import kieker.webgui.common.exception.ReflectionException;

import net.vidageek.mirror.dsl.Mirror;
import net.vidageek.mirror.exception.MirrorException;

/**
 * An analysis within the web gui.
 * 
 * @author Nils Christian Ehmke
 */
public class Analysis {

	private static final Log LOG = LogFactory.getLog(Analysis.class);

	private static final long MAX_THREAD_WAIT_TIME_MS = 1000;

	private final ClassAndMethodContainer classAndMethodContainer;
	private final Object analysisController;
	private final Object analysisControllerThread;

	private final Map<String, Map<String, Object>> displayPlugins = new HashMap<String, Map<String, Object>>();
	private final Map<String, Map<String, Object>> displayObjects = new HashMap<String, Map<String, Object>>();
	private final Map<String, Map<String, Method>> displayMethods = new HashMap<String, Map<String, Method>>();

	/**
	 * Creates a new instance of this class using the given parameters.
	 * 
	 * @param classLoader
	 *            The class loader to be used to initialize the analysis.
	 * @param projectFile
	 *            The file to be loaded.
	 * @throws InvalidAnalysisStateException
	 *             If something went wrong during the loading of the analysis.
	 * @throws AnalysisInitializationException
	 *             If an error occurred during the instantiation of the analysis.
	 */
	@SuppressWarnings("unchecked")
	public Analysis(final ClassLoader classLoader, final File projectFile) throws InvalidAnalysisStateException, AnalysisInitializationException {
		try {
			this.classAndMethodContainer = new ClassAndMethodContainer(classLoader);
			final Class<?> analysisControllerClass = this.classAndMethodContainer.getAnalysisControllerClass();
			final Class<?> analsisControllerThreadClass = this.classAndMethodContainer.getAnalysisControllerThreadClass();

			final Object modelProject = new Mirror().on(analysisControllerClass).invoke().method("loadFromFile").withArgs(projectFile.getAbsoluteFile());
			final Object controllerAndMapping = new Mirror().on(analysisControllerClass).invoke().method("createAnalysisController")
					.withArgs(modelProject, classLoader);

			this.analysisController = new Mirror().on(controllerAndMapping).invoke().method("getController").withoutArgs();
			this.analysisControllerThread = new Mirror().on(analsisControllerThreadClass).invoke().constructor().withArgs(this.analysisController);

			final Map<Object, Object> pluginMap = (Map<Object, Object>) new Mirror().on(controllerAndMapping).invoke().method("getPluginMap").withoutArgs();

			final List<Object> views = (List<Object>) new Mirror().on(modelProject).invoke().method("getViews").withoutArgs();

			for (final Object view : views) {
				final List<Object> displayConnectors = (List<Object>) new Mirror().on(view).invoke().method("getDisplayConnectors").withoutArgs();
				final String viewName = (String) new Mirror().on(view).invoke().method("getName").withoutArgs();

				if (!this.displayMethods.containsKey(viewName)) {
					this.displayMethods.put(viewName, new HashMap<String, Method>());
				}
				if (!this.displayObjects.containsKey(viewName)) {
					this.displayObjects.put(viewName, new HashMap<String, Object>());
				}
				if (!this.displayPlugins.containsKey(viewName)) {
					this.displayPlugins.put(viewName, new HashMap<String, Object>());
				}
				final Map<String, Method> methodMap = this.displayMethods.get(viewName);
				final Map<String, Object> displayObjectMap = this.displayObjects.get(viewName);
				final Map<String, Object> displayPluginMap = this.displayPlugins.get(viewName);

				for (final Object displayConnector : displayConnectors) {
					final Object display = new Mirror().on(displayConnector).invoke().method("getDisplay").withoutArgs();
					final String displayName = (String) new Mirror().on(display).invoke().method("getName").withoutArgs();
					final Object displayParent = new Mirror().on(display).invoke().method("getParent").withoutArgs();
					final Object plugin = pluginMap.get(displayParent);

					final Method[] methods = plugin.getClass().getMethods();
					for (final Method method : methods) {
						final Annotation displayAnnoation = method.getAnnotation(this.classAndMethodContainer.getDisplayAnnotationClass());
						if (displayAnnoation != null) {
							final String potentialDisplayName = (String) new Mirror().on(displayAnnoation).invoke().method("name").withoutArgs();
							if (displayName.equals(potentialDisplayName)) {
								methodMap.put(displayName, method);
								displayObjectMap.put(displayName, method.getParameterTypes()[0].newInstance());
								displayPluginMap.put(displayName, plugin);
								break;
							}
						}
					}
				}
			}

		} catch (final ReflectionException ex) {
			throw new AnalysisInitializationException("An error occured while instantiating the analysis.", ex);
		} catch (final NullPointerException ex) {
			throw new AnalysisInitializationException("An error occured while instantiating the analysis.", ex);
		} catch (final IllegalAccessException ex) {
			throw new AnalysisInitializationException("An error occured while instantiating the analysis.", ex);
		} catch (final InstantiationException ex) {
			throw new AnalysisInitializationException("An error occured while instantiating the analysis.", ex);
		} catch (final MirrorException ex) {
			throw new AnalysisInitializationException("An error occured while instantiating the analysis.", ex);
		}

	}

	/**
	 * Starts the analysis.
	 * 
	 * @throws InvalidAnalysisStateException
	 *             If the analysis is in the wrong state to be started.
	 */
	public void start() throws InvalidAnalysisStateException {
		try {
			new Mirror().on(this.analysisControllerThread).invoke().method("start").withoutArgs();
		} catch (final MirrorException ex) {
			Analysis.LOG.error("An error occured during a reflection method call.", ex);
		}
	}

	/**
	 * Stops the analysis.
	 */
	public void stop() {
		try {
			new Mirror().on(this.analysisControllerThread).invoke().method("terminate").withoutArgs();
			new Mirror().on(this.analysisControllerThread).invoke().method("join").withArgs(Analysis.MAX_THREAD_WAIT_TIME_MS);
		} catch (final MirrorException ex) {
			Analysis.LOG.error("An error occured during a reflection method call.", ex);
		}
	}

	/**
	 * Delivers the current state of the analysis.
	 * 
	 * @return The state.
	 */
	public Enum<?> getCurrentState() {
		try {
			return (Enum<?>) new Mirror().on(this.analysisController).invoke().method("getState").withoutArgs();
		} catch (final MirrorException ex) {
			Analysis.LOG.error("An error occured during a reflection method call.", ex);
			return null;
		}
	}

	/**
	 * This method delivers the display object of the (currently running) analysis. Technically it is an instance of {@code AbstractDisplay}, but in fact the project
	 * specific class loader has been used.
	 * 
	 * @param viewName
	 *            The name of the view.
	 * @param displayName
	 *            The name of the display.
	 * @return A display object for the given parameters.
	 */
	public Object getDisplay(final String viewName, final String displayName) {
		return this.displayObjects.get(viewName).get(displayName);
	}

	/**
	 * Delivers the log entries for the current analysis.
	 * 
	 * @return An array containing the log entries (if available).
	 */
	public Object[] getLogEntries() {
		try {
			return (Object[]) new Mirror().on(this.classAndMethodContainer.getLogImplWebguiLoggingClass()).invoke().method("getEntries").withoutArgs();
		} catch (final MirrorException ex) {
			return new Object[0];
		}
	}

	public void updateDisplays() {
		for (final String view : this.displayMethods.keySet()) {
			for (final String display : this.displayMethods.get(view).keySet()) {
				final Object obj = this.displayObjects.get(view).get(display);
				final Method method = this.displayMethods.get(view).get(display);
				final Object plugin = this.displayPlugins.get(view).get(display);

				new Mirror().on(plugin).invoke().method(method).withArgs(obj);
			}
		}
	}
}
