/***************************************************************************
 * Copyright 2013 Kieker Project (http://kieker-monitoring.net)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ***************************************************************************/

package kieker.webgui.web.beans.view;

import java.awt.Point;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.annotation.Nullable;
import javax.faces.application.Application;
import javax.faces.application.FacesMessage;
import javax.faces.component.UIComponent;
import javax.faces.component.UIInput;
import javax.faces.component.html.HtmlOutputText;
import javax.faces.context.FacesContext;
import javax.faces.event.AbortProcessingException;
import javax.faces.event.AjaxBehaviorEvent;

import com.google.common.base.Predicate;
import com.google.common.collect.Collections2;

import kieker.analysis.model.analysisMetaModel.MIAnalysisMetaModelFactory;
import kieker.analysis.model.analysisMetaModel.MIDisplay;
import kieker.analysis.model.analysisMetaModel.MIDisplayConnector;
import kieker.analysis.model.analysisMetaModel.MIPlugin;
import kieker.analysis.model.analysisMetaModel.MIProject;
import kieker.analysis.model.analysisMetaModel.MIView;
import kieker.analysis.model.analysisMetaModel.impl.MAnalysisMetaModelFactory;
import kieker.common.logging.Log;
import kieker.common.logging.LogFactory;
import kieker.monitoring.core.registry.Registry;
import kieker.webgui.common.exception.NewerProjectException;
import kieker.webgui.common.exception.ProjectLoadException;
import kieker.webgui.common.exception.ProjectNotExistingException;
import kieker.webgui.domain.ComponentListContainer;
import kieker.webgui.domain.pluginDecorators.AbstractPluginDecorator;
import kieker.webgui.domain.pluginDecorators.FilterDecorator;
import kieker.webgui.domain.pluginDecorators.ReaderDecorator;
import kieker.webgui.domain.pluginDecorators.RepositoryDecorator;
import kieker.webgui.service.IProjectService;
import kieker.webgui.web.beans.application.GlobalPropertiesBean;
import kieker.webgui.web.beans.application.ProjectsBean;
import kieker.webgui.web.beans.session.UserBean;

import org.primefaces.component.behavior.ajax.AjaxBehavior;
import org.primefaces.component.behavior.ajax.AjaxBehaviorListenerImpl;
import org.primefaces.component.dashboard.Dashboard;
import org.primefaces.component.panel.Panel;
import org.primefaces.context.RequestContext;
import org.primefaces.event.DashboardReorderEvent;
import org.primefaces.event.TabChangeEvent;
import org.primefaces.model.DashboardColumn;
import org.primefaces.model.DashboardModel;
import org.primefaces.model.DefaultDashboardColumn;
import org.primefaces.model.DefaultDashboardModel;

import org.eclipse.emf.common.util.EList;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;

/**
 * The {@link CurrentCockpitEditorBean} contains the necessary data behind an instance of the cockpit editor.
 * The class is a Spring managed bean with view scope to make sure that one user (even in one session) can open multiple projects at a time without causing any
 * problems.
 * 
 * @author Nils Christian Ehmke
 */
@Component
@Scope("view")
public class CurrentCockpitEditorBean {

	private static final int NUMBER_COLUMNS = 2;
	private static final Log LOG = LogFactory.getLog(CurrentCockpitEditorBean.class);

	private final MIAnalysisMetaModelFactory factory = MAnalysisMetaModelFactory.eINSTANCE;
	@Autowired
	private IProjectService projectService;

	private ComponentListContainer availableComponents;
	private long currId;
	private long timeStamp;
	private String projectName;
	private MIProject project;
	private boolean unsavedModifications;
	private MIView activeView;
	private Dashboard dashboard;
	private DashboardModel dashboardModel;
	@Autowired
	private ProjectsBean projectsBean;
	@Autowired
	private GlobalPropertiesBean globalPropertiesBean;
	@Autowired
	private UserBean userBean;

	private final Registry<MIDisplayConnector> connectors = new Registry<MIDisplayConnector>();
	private MIDisplayConnector selectedNode = null;

	/**
	 * Creates a new instance of this class. <b>Do not call this constructor manually. It will only be accessed by Spring.</b>
	 */
	public CurrentCockpitEditorBean() {
		this.availableComponents = new ComponentListContainer(Collections.<ReaderDecorator>emptyList(), Collections.<FilterDecorator>emptyList(),
				Collections.<RepositoryDecorator>emptyList());
		this.createDashboard();
	}

	/**
	 * Delivers the currently selected display connector.
	 * 
	 * @return The selected connector.
	 */
	public MIDisplayConnector getSelectedNode() {
		return this.selectedNode;
	}

	/**
	 * Creates the initial dashboard object.
	 */
	private void createDashboard() {
		final FacesContext fc = FacesContext.getCurrentInstance();
		final Application application = fc.getApplication();

		// Create the primefaces dashboard
		this.dashboard = (Dashboard) application.createComponent(fc, "org.primefaces.component.Dashboard", "org.primefaces.component.DashboardRenderer");
		this.dashboard.setId("dashboard");

		// Create the model and add the columns
		this.dashboardModel = new DefaultDashboardModel();
		for (int i = 0; i < CurrentCockpitEditorBean.NUMBER_COLUMNS; i++) {
			final DashboardColumn column = new DefaultDashboardColumn();
			this.dashboardModel.addColumn(column);
		}
		this.dashboard.setModel(this.dashboardModel);
	}

	public boolean isUnsavedModification() {
		return this.unsavedModifications;
	}

	/**
	 * Fills the initial dashboard object.
	 */
	private void fillDashboard() {
		// Dump the old entries
		this.clearDashboard();

		// Now add the entries from the current view
		if (this.activeView != null) {
			// Add a panel for every display connector we have
			this.currId = 0;

			for (int col = 0; col < 2; col++) {
				final Collection<MIDisplayConnector> displayConnectors = this.getSortedDisplayConnectors(this.activeView.getDisplayConnectors(), col);
				final DashboardColumn column = this.dashboard.getModel().getColumn(col);

				for (final MIDisplayConnector connector : displayConnectors) {
					final Panel panel = this.createPanelFromDisplayConnector(connector);

					this.dashboard.getChildren().add(panel);
					column.addWidget(panel.getId());

					this.currId++;
				}
			}
		}
	}

	private Collection<MIDisplayConnector> getSortedDisplayConnectors(final EList<MIDisplayConnector> eList, final int col) {
		final List<MIDisplayConnector> l = new ArrayList<MIDisplayConnector>(Collections2.filter(eList, new ColumnPredicate(col)));

		Collections.sort(l, new PosComparator());

		return l;
	}

	/**
	 * This method initializes the bean by using the current project name to load the project. <b>Do not call this method manually. It will only be accessed by
	 * Spring.</b>
	 */
	public void initalize() {

		try {
			// Make sure that the initialization will only be done for the init request.
			if (!FacesContext.getCurrentInstance().isPostback()) {
				// Remember the given parameters
				this.project = this.projectsBean.openProject(this.projectName);
				if (this.project != null) {
					// Remember the current time! This is important for the later comparison of the time stamps.
					this.resetTimeStamp();
					this.reloadComponents();

					this.stringToLayout(this.projectService.getCockpitLayout(this.projectName));

					// Update the class loader and the specific classes used within various methods in this bean
					this.fillDashboard();

					if (!this.project.getViews().isEmpty()) {
						this.setActiveView(this.project.getViews().get(0));
					}
				}

				this.unsavedModifications = false;
			}
		} catch (final ProjectLoadException ex) {
			CurrentCockpitEditorBean.LOG.error("An error occured while loading the project.", ex);
			GlobalPropertiesBean.showMessage(FacesMessage.SEVERITY_ERROR, "An error occured while loading the project.");
		} catch (final NullPointerException ex) {
			CurrentCockpitEditorBean.LOG.error("An error occured while loading the project.", ex);
			GlobalPropertiesBean.showMessage(FacesMessage.SEVERITY_ERROR, "An error occured while loading the project.");
		}
	}

	private void reloadComponents() {
		this.availableComponents = this.projectService.getAvailableComponents(this.projectName);
	}

	public MIProject getProject() {
		return this.project;
	}

	/**
	 * This is a dummy method returning just a collection of null objects. This is necessary due to Primefaces.
	 * 
	 * @return A collection with two null objects.
	 */
	public Collection<Object> getProperties() {
		return Collections.nCopies(3, null);
	}

	public void setProjectName(final String newName) {
		this.projectName = newName;
	}

	public String getProjectName() {
		return this.projectName;
	}

	/**
	 * This method can be used to get the description of a {@link MIDisplay}. Currently it is a little bit expensive to search for the description.
	 * 
	 * @param display
	 *            The display whose description should be extracted.
	 * @return The description for the display or a substitute if none is available. This is in either case human readable.
	 */
	public String getDescription(final MIDisplay display) {
		final String parentClassname = display.getParent().getClassname();
		AbstractPluginDecorator<? extends MIPlugin> parentContainer = null;

		// Find the correct parent container
		for (final FilterDecorator plugin : this.availableComponents.getFilters()) {
			if (plugin.getClassname().equals(parentClassname)) {
				parentContainer = plugin;
				break;
			}
		}
		if (parentContainer == null) {
			for (final ReaderDecorator plugin : this.availableComponents.getReaders()) {
				if (plugin.getClassname().equals(parentClassname)) {
					parentContainer = plugin;
					break;
				}
			}
		}

		// If we have now the correct parent, we can search for the correct display instance.
		if (parentContainer != null) {
			return parentContainer.getDisplayDescription(display.getName());
		}

		// Nothing found
		return "No description available";
	}

	/**
	 * This method tries to save the current project and informs the user about success or fail.
	 * 
	 * @param overwriteNewerProject
	 *            This flag determines whether a newer project should be overwritten.
	 */
	public void saveProject(final boolean overwriteNewerProject) {
		try {
			this.projectService.saveProject(this.projectName, this.project, this.timeStamp, overwriteNewerProject, this.userBean.getUsername(), null,
					this.layoutToString());
			GlobalPropertiesBean.showMessage(FacesMessage.SEVERITY_INFO, this.globalPropertiesBean.getMsgProjectSaved());
			// Update the time stamp!
			this.resetTimeStamp();

			this.clearModificationsFlag();
		} catch (final IOException ex) {
			CurrentCockpitEditorBean.LOG.error("An error occured while saving the projet.", ex);
			GlobalPropertiesBean.showMessage(FacesMessage.SEVERITY_ERROR, "An error occured while saving the project.");
		} catch (final NewerProjectException ex) {
			CurrentCockpitEditorBean.LOG.info("The project has been modified externally in the meanwhile.", ex);
			GlobalPropertiesBean.showMessage(FacesMessage.SEVERITY_WARN, "The project has been modified externally in the meanwhile.");
			// Give the user the possibility to force-save the project
			RequestContext.getCurrentInstance().execute("forceSaveDlg.show()");
		} catch (final ProjectNotExistingException ex) {
			CurrentCockpitEditorBean.LOG.error("A project with the given name does not exist.", ex);
			GlobalPropertiesBean.showMessage(FacesMessage.SEVERITY_ERROR, "A project with the given name does not exist.");
		}
	}

	/**
	 * This method sets the property {@link CurrentCockpitEditorBean#unsavedModifications} to false and refreshes the necessary components within the editor to make
	 * this visible.
	 */
	private void clearModificationsFlag() {
		this.unsavedModifications = false;
		RequestContext.getCurrentInstance().update("menuForm");
	}

	/**
	 * This method sets the property {@link CurrentCockpitEditorBean#unsavedModifications} to true and refreshes the necessary components within the editor
	 * to make this visible.
	 */
	private void setModificationsFlag() {
		this.unsavedModifications = true;
		RequestContext.getCurrentInstance().update("menuForm");
	}

	/**
	 * This method sets the time stamp to the current system time.
	 */
	public void resetTimeStamp() {
		this.timeStamp = System.currentTimeMillis();
	}

	/**
	 * This method adds a new view to the project.
	 * 
	 * @param viewName
	 *            The name of the new view.
	 */
	public void addView(final String viewName) {
		if (this.project != null) {
			// Create the view and add it to our project
			final MIView view = this.factory.createView();
			view.setName(viewName);
			view.setDescription("No description available.");
			this.project.getViews().add(view);

			this.setModificationsFlag();
		}
	}

	public MIView getActiveView() {
		return this.activeView;
	}

	/**
	 * Sets the active view and updates the dashboard.
	 * 
	 * @param view
	 *            The new view.
	 */
	public void setActiveView(final MIView view) {
		this.activeView = view;

		this.fillDashboard();
	}

	/**
	 * Deletes the given view from the model.
	 * 
	 * @param view
	 *            The view to be removed.
	 * 
	 */
	public void deleteView(final MIView view) {
		this.project.getViews().remove(view);

		this.setModificationsFlag();
	}

	private Panel createPanelFromDisplayConnector(final MIDisplayConnector connector) {
		final FacesContext fc = FacesContext.getCurrentInstance();
		final Application application = fc.getApplication();
		final Panel panel = (Panel) application.createComponent(fc, "org.primefaces.component.Panel", "org.primefaces.component.PanelRenderer");
		final String id = this.displayConnectorToID(connector);

		// Set the usual properties of the panel
		panel.setId(id);
		panel.setHeader(connector.getName());
		panel.setClosable(true);
		panel.setToggleable(false);

		final HtmlOutputText text = new HtmlOutputText();
		text.setValue(connector.getDisplay().getName());
		panel.getChildren().add(text);

		// The following code makes sure that the application detects the close event
		final AjaxBehavior behaviour = new AjaxBehavior();
		behaviour.setProcess("@this");
		behaviour.addAjaxBehaviorListener(new AjaxBehaviorListenerImpl() {

			private static final long serialVersionUID = 1L;

			@Override
			public void processAjaxBehavior(final AjaxBehaviorEvent event) throws AbortProcessingException {
				CurrentCockpitEditorBean.this.panelCloseEvent(event);
			}

		});
		panel.addClientBehavior("close", behaviour);

		return panel;
	}

	private void panelCloseEvent(final AjaxBehaviorEvent event) {
		if (CurrentCockpitEditorBean.this.activeView != null) {
			final String id = ((Panel) event.getSource()).getId();
			final MIDisplayConnector connector = CurrentCockpitEditorBean.this.idToDisplayConnector(id);

			CurrentCockpitEditorBean.this.activeView.getDisplayConnectors().remove(connector);
			CurrentCockpitEditorBean.this.saveLayoutOfCurrentView();
		}
	}

	/**
	 * This method adds the given display to the currently active view. If no view exists, this method does nothing.
	 * 
	 * @param display
	 *            The display which should be added to the current view.
	 */
	public void addDisplayToView(final MIDisplay display) {
		if (this.activeView != null) {
			final MIDisplayConnector connector = this.factory.createDisplayConnector();
			connector.setDisplay(display);
			connector.setName("Display " + this.currId);
			this.activeView.getDisplayConnectors().add(connector);

			// Now add it to the dashboard as well
			final Panel panel = this.createPanelFromDisplayConnector(connector);
			this.getDashboard().getChildren().add(panel);
			final DashboardColumn column = this.dashboardModel.getColumn(0);
			column.addWidget(panel.getId());
			this.currId++;

			this.setModificationsFlag();

			this.saveLayoutOfCurrentView();
		}
	}

	private static final String ID_PREFIX = "displayConnector_";

	final Map<MIView, Map<MIDisplayConnector, Point>> connectorPositions = new HashMap<MIView, Map<MIDisplayConnector, Point>>();

	public void handleReorder(final DashboardReorderEvent event) {
		this.saveLayoutOfCurrentView();
	}

	/**
	 * This method is used as a validator for new display connector names.
	 * 
	 * @param context
	 *            The context of the validation.
	 * @param toValidate
	 *            The components which has be validated.
	 * @param value
	 *            The new value.
	 */
	public void validateDisplayConnectorName(final FacesContext context, final UIComponent toValidate, final Object value) {
		if ((value instanceof String) && (toValidate instanceof UIInput)) {
			final boolean nameExists = this.existsDisplayConnectorName((String) value);
			((UIInput) toValidate).setValid(!nameExists);
		}
	}

	/**
	 * This event is used if the currently selected tab changes. It modifies the active view property within this bean.
	 * 
	 * @param event
	 *            The onChange-event.
	 */
	public void onChange(final TabChangeEvent event) {
		if (event.getData() instanceof MIView) {
			this.setActiveView((MIView) event.getData());
		}
	}

	/**
	 * This is the event if a node has been clicked and should be selected.
	 */
	public void nodeSelected() {
		final Map<String, String> paramMap = FacesContext.getCurrentInstance().getExternalContext().getRequestParameterMap();

		final String fullID = paramMap.get("id");
		final String shortID = fullID.substring(fullID.indexOf(':') + "displayConnector_".length() + 1);

		final MIDisplayConnector connector = this.connectors.get(Integer.parseInt(shortID));
		if (connector != null) {
			this.selectedNode = connector;
		}
	}

	/**
	 * Getter for the property {@link CurrentCockpitEditorBean#dashboard}.
	 * 
	 * @return The current value of the property.
	 */
	public Dashboard getDashboard() {
		return this.dashboard;
	}

	/**
	 * Setter for the property {@link CurrentCockpitEditorBean#dashboard}.
	 * 
	 * @param dashboard
	 *            The new value for the property.
	 */
	public void setDashboard(final Dashboard dashboard) {
		this.dashboard = dashboard;
		this.dashboard.setModel(this.dashboardModel);
	}

	/**
	 * This method checks whether a display connector with the given name exists already.
	 * 
	 * @param name
	 *            The name to be checked.
	 * @return true iff the name exists already.
	 */
	private boolean existsDisplayConnectorName(final String name) {
		// Make sure a view is selected
		if (this.activeView == null) {
			return false;
		}

		boolean result = false;
		// Run through all display connectors and check the name against the given one
		for (final MIDisplayConnector connector : this.activeView.getDisplayConnectors()) {
			if (connector.getName().equals(name)) {
				result = true;
				break;
			}
		}

		// The name has not been found
		return result;
	}

	private void clearDashboard() {
		// Run through all columns of the dashboard model and remove the items
		final List<DashboardColumn> columns = this.dashboard.getModel().getColumns();
		for (final DashboardColumn column : columns) {
			column.getWidgets().clear();
		}
		// Now remove the items from the dashboard
		this.dashboard.getChildren().clear();
	}

	private void saveLayoutOfCurrentView() {
		if (this.activeView != null) {
			final Map<MIDisplayConnector, Point> layout = new HashMap<MIDisplayConnector, Point>();

			// Run through all columns and put the position of every widget into the map
			int col = 0;
			for (final DashboardColumn column : this.dashboardModel.getColumns()) {
				int row = 0;
				for (final String widgetID : column.getWidgets()) {
					layout.put(this.idToDisplayConnector(widgetID), new Point(col, row));
					row++;
				}
				col++;
			}

			// Store the map for the whole view
			this.connectorPositions.put(this.activeView, layout);
		}
	}

	private void stringToLayout(final String layout) {
		final String[] elements = layout.split(" ");
		int i = 0;

		for (final MIView view : this.project.getViews()) {
			final Map<MIDisplayConnector, Point> positions = new HashMap<MIDisplayConnector, Point>();
			this.connectorPositions.put(view, positions);
			for (final MIDisplayConnector connector : view.getDisplayConnectors()) {
				final Point pos = new Point(Integer.valueOf(elements[i]), Integer.valueOf(elements[i + 1]));

				positions.put(connector, pos);

				i += 2;
			}
		}

	}

	private String layoutToString() {
		final StringBuilder builder = new StringBuilder();

		for (final MIView view : this.project.getViews()) {
			final Map<MIDisplayConnector, Point> positions = this.connectorPositions.get(view);
			for (final MIDisplayConnector connector : view.getDisplayConnectors()) {
				final Point pos = positions.get(connector);

				builder.append(pos.x);
				builder.append(' ');
				builder.append(pos.y);
				builder.append(' ');
			}
		}

		return builder.toString();
	}

	private String displayConnectorToID(final MIDisplayConnector displayConnector) {
		return ID_PREFIX + this.connectors.get(displayConnector);
	}

	private MIDisplayConnector idToDisplayConnector(final String id) {
		final String shortID = id.substring(ID_PREFIX.length());
		final int intID = Integer.valueOf(shortID);

		return this.connectors.get(intID);
	}

	private class PosComparator implements Comparator<MIDisplayConnector> {

		@Override
		public int compare(final MIDisplayConnector o1, final MIDisplayConnector o2) {
			final int pos1 = CurrentCockpitEditorBean.this.connectorPositions.get(CurrentCockpitEditorBean.this.activeView).get(o1).y;
			final int pos2 = CurrentCockpitEditorBean.this.connectorPositions.get(CurrentCockpitEditorBean.this.activeView).get(o2).y;

			return pos1 - pos2;
		}

	}

	private class ColumnPredicate implements Predicate<MIDisplayConnector> {

		private final int col;

		public ColumnPredicate(final int col) {
			this.col = col;
		}

		@Override
		public boolean apply(@Nullable final MIDisplayConnector key) {
			return CurrentCockpitEditorBean.this.connectorPositions.get(CurrentCockpitEditorBean.this.activeView).get(key).x == this.col;
		}

	}

}
